;Low-power LCD clock with DCF77 reception
;ATMega8515 @ 32768 Hz (external crystal)
;
;PORTS:
;  PA0-7 = LCD segments { g1, f1, a1, b1,  g2, f2, a2, b2    }
;  PC0-7 = LCD segments { e1, d1, c1, dp1, d2, d2, c2, dp2   }
;  PC0-7 = LCD segments { a4, f4, g4, b3,  a3, f3, g3, colon }
;  PD0-7 = LCD segments { e3, d3, c3, dp3, e4, d4, c4, b4    }
;  PE0   = DCF77 input
;  PE1   = DCF77 enable output
;  PE2   = LCD backplane
;
;VERSIONS
;  2011-05-29  Arne Rossius
;    * first version

.include "../include/m8515def.inc"

.def	sreg_backup = R3
.def	data_a = R4
.def	data_b = R5
.def	data_c = R6
.def	data_d = R7
.def	out_a = R8
.def	out_b = R9
.def	out_c = R10
.def	out_d = R11
.def	digit1 = R12
.def	digit2 = R13
.def	digit3 = R14
.def	digit4 = R15

.def	temp = R16
.def	temp2 = R17
.def	temp3 = R18
.def	temp4 = R19
.def	inttemp = R20

.def	dcf_timeout = R22
.def	dcfbits = R23
.def	dcfflags = R24
.def	flags = R25
	.equ	fBackplane = 0
	.equ	fIncrementClock = 1
	.equ	fUpdateDisplay = 2
	.equ	fInitialReception = 3
	.equ	fNoReception = 4
.def	subsec = R26
.def	seconds = R27
;R29:R28 = Y pointer

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org OC0addr
	rjmp	oc0
	
.equ DCF_ENABLE_INVERTED = 0 ;TODO invert DCF77 enable output (1 = active low)
.equ DCF_PULLUP = 1 ;enable internal pull-up for DCF77 input
.equ DCF_INVERTED = 1 ;DCF77 receiver with inverted output (1 = negative pulses)
.equ DCF_USE_INT0 = 0
.equ DCF_USE_UTC = 0
.include "dcf77.asm"

segment_table:
	;digits '0' to '9' for 7-segment display
	.db 0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07, 0x7F, 0x6F
	
;===============================================================================

oc0:
	;64 Hz interrupt
	in	sreg_backup, SREG
	
	;output current LCD data
	in	inttemp, PORTE
	andi	inttemp, ~(0x04)
	sbrc	flags, fBackplane
	ori	inttemp, 0x04
	out	PORTA, out_a
	out	PORTB, out_b
	out	PORTE, inttemp
	out	PORTC, out_c
	out	PORTD, out_d
	
	;toggle segments and backplane
	movw	out_b:out_a, data_b:data_a
	movw	out_d:out_c, data_d:data_c
	ldi	inttemp, 1<<fBackplane
	eor	flags, inttemp
	sbrs	flags, fBackplane
	rjmp	oc0_toggle_end
	com	out_a
	com	out_b
	com	out_c
	com	out_d
oc0_toggle_end:
	
	;increment clock ('seconds' doesn't contain real seconds!)
	adiw	seconds:subsec, 16
	cpi	seconds, 0xF0 ;16*60*64 = 0xF000
	brlo	oc0_inc_end
	clr	seconds
	clr	subsec
	ori	flags, 1<<fIncrementClock
oc0_inc_end:

oc0_end:
	out	SREG, sreg_backup
	reti

;===============================================================================

reset:
	;set stackpointer
	ldi	temp, HIGH(RAMEND)
	out	SPH, temp
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;init ports
	ldi	temp, 0xFF
	out	DDRA, temp
	out	DDRB, temp
	out	DDRC, temp
	out	DDRD, temp
	ldi	temp, 0x06
	out	DDRE, temp
	ldi	temp, 0x00
	out	PORTA, temp
	out	PORTB, temp
	out	PORTC, temp
	out	PORTD, temp
	.if (DCF_PULLUP == 0)
	ldi	temp, 0x02
	.else
	ldi	temp, 0x03
	.endif
	out	PORTE, temp
	
	;disable analog comparator
	ldi	temp, 1<<ACD
	out	ACSR, temp
	
	;init timer 0 (LCD update interrupt)
	ldi	temp, 1<<WGM01 | 0x02 ;CTC mode, Clk/8
	out	TCCR0, temp
	ldi	temp, 63
	out	OCR0, temp
	ldi	temp, 1<<TOIE0 | 1<<OCIE0
	out	TIMSK, temp
	
	;init registers
	clr	out_a
	clr	out_b
	clr	out_c
	clr	out_d
	clr	data_a
	clr	data_b
	clr	data_c
	clr	data_d
	clr	subsec
	clr	seconds
	clr	dcf_timeout
	ldi	flags, 0
	
	;init DCF77 receiver
	rcall	dcf_init
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------

loop:
	;increment time?
	sbrs	flags, fIncrementClock
	rjmp	increment_clock_end
	andi	flags, ~(1<<fIncrementClock)
	cpi	dcf_timeout, 0
	breq	PC+2
	dec	dcf_timeout
	sbrc	dcfflags, dcfTime ;don't increment if time just set by DCF77
	rjmp	increment_clock_skip
	ldi	YL, LOW(RAM_Time)
	ldi	YH, HIGH(RAM_Time)
	rcall	inctime
increment_clock_skip:
	ori	flags, 1<<fUpdateDisplay
	;start DCF77 reception?
	lds	temp, RAM_Time+tHour
	cpi	temp, 0x03
	brne	increment_clock_end
	lds	temp, RAM_Time+tMinute
	cpi	temp, 0x00
	brne	increment_clock_end
	;start DCF77 reception
	sbis	PORTE, 1
	ori	flags, 1<<fNoReception
	sbi	PORTE, 1
	ldi	dcf_timeout, 15
increment_clock_end:

	;new time received?
	sbrs	dcfflags, dcfTime
	rjmp	new_time_end
	andi	dcfflags, ~(1<<dcfTime)
	andi	flags, ~(1<<fNoReception)
	ori	flags, 1<<fUpdateDisplay | 1<<fInitialReception
	clr	dcf_timeout
new_time_end:

	;stop DCF77 reception?
	cpi	dcf_timeout, 0
	brne	stop_dcf_end
	sbrs	flags, fInitialReception
	rjmp	stop_dcf_end
;	sbis	PORTE, 1
;	rjmp	stop_dcf_end
	cbi	PORTE, 1
stop_dcf_end:
	
	;update display?
	sbrs	flags, fUpdateDisplay
	rjmp	update_display_end
	andi	flags, ~(1<<fUpdateDisplay)
	ldi	temp2, 0
	;digit 1
	ldi	ZL, LOW(segment_table*2)
	ldi	ZH, HIGH(segment_table*2)
	lds	temp, RAM_Time+tHour
	swap	temp
	andi	temp, 0x0F
	add	ZL, temp
	adc	ZH, temp2
	lpm	temp, Z+
	cpi	temp, 0x3F ;suppress leading zero
	brne	PC+2
	ldi	temp, 0x00
	mov	digit1, temp
	;digit 2
	ldi	ZL, LOW(segment_table*2)
	ldi	ZH, HIGH(segment_table*2)
	lds	temp, RAM_Time+tHour
	andi	temp, 0x0F
	add	ZL, temp
	adc	ZH, temp2
	lpm	digit2, Z+
	;digit 3
	ldi	ZL, LOW(segment_table*2)
	ldi	ZH, HIGH(segment_table*2)
	lds	temp, RAM_Time+tMinute
	swap	temp
	andi	temp, 0x0F
	add	ZL, temp
	adc	ZH, temp2
	lpm	digit3, Z+
	;digit 4
	ldi	ZL, LOW(segment_table*2)
	ldi	ZH, HIGH(segment_table*2)
	lds	temp, RAM_Time+tMinute
	andi	temp, 0x0F
	add	ZL, temp
	adc	ZH, temp2
	lpm	digit4, Z+
	rcall	convert_digits
update_display_end:
	
	;display DCF77 reception info?
	sbrc	flags, fInitialReception
	rjmp	display_dcfinfo_end
	;digits 1 and 2: number of received bits
	mov	temp, dcfbits
	clr	temp2
display_dcfinfo_bin2bcd:
	cpi	temp, 10
	brlo	display_dcfinfo_bin2bcd_end
	subi	temp, 10
	inc	temp2
	rjmp	display_dcfinfo_bin2bcd
display_dcfinfo_bin2bcd_end:
	ldi	ZL, LOW(segment_table*2)
	ldi	ZH, HIGH(segment_table*2)
	add	ZL, temp2
	ldi	temp2, 0
	adc	ZH, temp2
	lpm	digit1, Z+
	ldi	ZL, LOW(segment_table*2)
	ldi	ZH, HIGH(segment_table*2)
	add	ZL, temp
	adc	ZH, temp2
	lpm	temp, Z+
	in	temp2, TCCR1B
	.if (DCF_INVERTED == 0)
	sbrs	temp2, ICES1
	.else
	sbrc	temp2, ICES1
	.endif
	ori	temp, 0x80 ;dp2
	mov	digit2, temp
	;digit 3: last error
	ldi	temp, 0x00 ;' '
	sbrc	dcfflags, dcfFirstTelegram
	ldi	temp, 0x06 ;'1' = one telegram received sucessfully
	sbrc	dcfflags, dcfStop
	ldi	temp, 0x40 ;'-' = waiting (DCF stopped, but no error)
	mov	temp2, dcfflags
	andi	temp2, dcfErrMask
	cpi	temp2, dcfErrBitLength
	brne	PC+2
	ldi	temp, 0x38 ;'L' = Bit Length error
	cpi	temp2, dcfErrBitCount
	brne	PC+2
	ldi	temp, 0x39 ;'C' = Bit Count error
	cpi	temp2, dcfErrParity
	brne	PC+2
	ldi	temp, 0x73 ;'P' = Parity error
	cpi	temp2, dcfErrStartBit
	brne	PC+2
	ldi	temp, 0x5D ;'S' = Start Bit error
	cpi	temp2, dcfErrDST
	brne	PC+2
	ldi	temp, 0x5E ;'d' = invalid DST bits
	cpi	temp2, dcfErrPlausibility
	brne	PC+2
	ldi	temp, 0x79 ;'E' = Plausibility Check failed
	mov	digit3, temp
	;digit 4: last received bit
	ldi	temp, 0x40 ;'-'
	cpi	dcfbits, 0
	breq	display_dcfinfo_digit4_end
	sbrc	dcfflags, dcfStop
	rjmp	display_dcfinfo_digit4_end
	ldi	temp, 0x3F ;'0'
	lds	temp2, RAM_DCF2+5
	sbrc	temp2, 7
	ldi	temp, 0x06 ;'1'
display_dcfinfo_digit4_end:
	mov	digit4, temp
	rcall	convert_digits
display_dcfinfo_end:

	;display flashing/static colon or DCF77 reception indicator (dp2)
	sbic	PORTE, 1
	rjmp	colon_rx
	bst	seconds, 1
	sbrs	flags, fNoReception
	set ;static colon if last reception was successful
	bld	data_c, 7
	rjmp	colon_end
colon_rx:
	;DCF77 reception indicator
	sbrs	flags, fInitialReception
	rjmp	colon_end
	in	temp, TCCR1B
	clt
	.if (DCF_INVERTED == 0)
	sbrs	temp, ICES1
	.else
	sbrc	temp, ICES1
	.endif
	set
	bld	data_b, 7
colon_end:

	;sleep (idle mode)
	in	temp, MCUCR
	ori	temp, 1<<SE
	out	MCUCR, temp
	sei
	sleep
	
	rjmp	loop

;===============================================================================

convert_digits:
	;convert per-digit data to (non-inverted) port output data
	;digit 1
	clr	temp
	clr	temp2
	clr	temp3
	clr	temp4
	sbrc	digit1, 0 ;a
	ori	temp, 0x04
	sbrc	digit1, 1 ;b
	ori	temp, 0x08
	sbrc	digit1, 2 ;c
	ori	temp2, 0x04
	sbrc	digit1, 3 ;d
	ori	temp2, 0x02
	sbrc	digit1, 4 ;e
	ori	temp2, 0x01
	sbrc	digit1, 5 ;f
	ori	temp, 0x02
	sbrc	digit1, 6 ;g
	ori	temp, 0x01
	sbrc	digit1, 7 ;dp
	ori	temp2, 0x08
	;digit 2
	sbrc	digit2, 0 ;a
	ori	temp, 0x40
	sbrc	digit2, 1 ;b
	ori	temp, 0x80
	sbrc	digit2, 2 ;c
	ori	temp2, 0x40
	sbrc	digit2, 3 ;d
	ori	temp2, 0x20
	sbrc	digit2, 4 ;e
	ori	temp2, 0x10
	sbrc	digit2, 5 ;f
	ori	temp, 0x20
	sbrc	digit2, 6 ;g
	ori	temp, 0x10
	sbrc	digit2, 7 ;dp
	ori	temp2, 0x80
	;digit 3
	sbrc	digit3, 0 ;a
	ori	temp3, 0x10
	sbrc	digit3, 1 ;b
	ori	temp3, 0x08
	sbrc	digit3, 2 ;c
	ori	temp4, 0x04
	sbrc	digit3, 3 ;d
	ori	temp4, 0x02
	sbrc	digit3, 4 ;e
	ori	temp4, 0x01
	sbrc	digit3, 5 ;f
	ori	temp3, 0x20
	sbrc	digit3, 6 ;g
	ori	temp3, 0x40
	sbrc	digit3, 7 ;dp
	ori	temp4, 0x08
	;digit 4
	sbrc	digit4, 0 ;a
	ori	temp3, 0x01
	sbrc	digit4, 1 ;b
	ori	temp4, 0x80
	sbrc	digit4, 2 ;c
	ori	temp4, 0x40
	sbrc	digit4, 3 ;d
	ori	temp4, 0x20
	sbrc	digit4, 4 ;e
	ori	temp4, 0x10
	sbrc	digit4, 5 ;f
	ori	temp3, 0x02
	sbrc	digit4, 6 ;g
	ori	temp3, 0x04
	sbrc	digit4, 7 ;colon
	ori	temp3, 0x80
	
	movw	data_b:data_a, temp2:temp
	movw	data_d:data_c, temp4:temp3
	
	ret
